@@ -1,6 +1,6 @@
 /* horst - Highly Optimized Radio Scanning Tool
  *
- * Copyright (C) 2005-2011 Bruno Randolf (br1@einfach.org)
+ * Copyright (C) 2005-2014 Bruno Randolf (br1@einfach.org)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -19,6 +19,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <unistd.h>
 #include <string.h>
 #include <getopt.h>
 #include <signal.h>
@@ -33,16 +34,18 @@
 #include "protocol_parser.h"
 #include "network.h"
 #include "display.h"
-#include "ieee80211.h"
+#include "wlan_util.h"
 #include "ieee80211_util.h"
-#include "wext.h"
+#include "control.h"
+#include "channel.h"
+#include "node.h"
+#include "essid.h"
 
 
 struct list_head nodes;
 struct essid_meta_info essids;
 struct history hist;
 struct statistics stats;
-struct chan_freq channels[MAX_CHANNELS];
 struct channel_info spectrum[MAX_CHANNELS];
 
 struct config conf = {
@@ -50,9 +53,11 @@ struct config conf = {
 	.channel_time		= CHANNEL_TIME,
 	.ifname			= INTERFACE_NAME,
 	.display_interval	= DISPLAY_UPDATE_INTERVAL,
-	.filter_pkt		= 0xffffff,
 	.recv_buffer_size	= RECV_BUFFER_SIZE,
 	.port			= DEFAULT_PORT,
+	.control_pipe		= DEFAULT_CONTROL_PIPE,
+	.filter_pkt		= PKT_TYPE_ALL,
+	.filter_mode		= WLAN_MODE_ALL,
 };
 
 struct timeval the_time;
@@ -90,13 +95,6 @@ static fd_set excpt_fds;
 static struct timeval tv;
 
 
-struct node_info* node_update(struct packet_info* p);
-void update_essids(struct packet_info* p, struct node_info* n);
-void timeout_nodes(void);
-int auto_change_channel(int mon);
-void get_current_channel(int mon);
-
-
 void __attribute__ ((format (printf, 1, 2)))
 printlog(const char *fmt, ...)
 {
@@ -126,7 +124,7 @@ update_history(struct packet_info* p)
 	hist.signal[hist.index] = p->phy_signal;
 	hist.noise[hist.index] = p->phy_noise;
 	hist.rate[hist.index] = p->phy_rate;
-	hist.type[hist.index] = p->wlan_type;
+	hist.type[hist.index] = (p->phy_flags & PHY_FLAG_BADFCS) ? 1 : p->wlan_type;
 	hist.retry[hist.index] = p->wlan_retry;
 
 	hist.index++;
@@ -138,7 +136,9 @@ update_history(struct packet_info* p)
 static void
 update_statistics(struct packet_info* p)
 {
-	if (p->phy_rate == 0)
+	int type = (p->phy_flags & PHY_FLAG_BADFCS) ? 1 : p->wlan_type;
+
+	if (p->phy_rate_idx == 0)
 		return;
 
 	stats.packets++;
@@ -146,17 +146,18 @@ update_statistics(struct packet_info* p)
 	if (p->wlan_retry)
 		stats.retries++;
 
-	if (p->phy_rate > 0 && p->phy_rate < MAX_RATES) {
+	if (p->phy_rate_idx > 0 && p->phy_rate_idx < MAX_RATES) {
 		stats.duration += p->pkt_duration;
-		stats.packets_per_rate[p->phy_rate]++;
-		stats.bytes_per_rate[p->phy_rate] += p->wlan_len;
-		stats.duration_per_rate[p->phy_rate] += p->pkt_duration;
+		stats.packets_per_rate[p->phy_rate_idx]++;
+		stats.bytes_per_rate[p->phy_rate_idx] += p->wlan_len;
+		stats.duration_per_rate[p->phy_rate_idx] += p->pkt_duration;
 	}
-	if (p->wlan_type >= 0 && p->wlan_type < MAX_FSTYPE) {
-		stats.packets_per_type[p->wlan_type]++;
-		stats.bytes_per_type[p->wlan_type] += p->wlan_len;
-		if (p->phy_rate > 0 && p->phy_rate < MAX_RATES)
-			stats.duration_per_type[p->wlan_type] += p->pkt_duration;
+
+	if (type >= 0 && type < MAX_FSTYPE) {
+		stats.packets_per_type[type]++;
+		stats.bytes_per_type[type] += p->wlan_len;
+		if (p->phy_rate_idx > 0 && p->phy_rate_idx < MAX_RATES)
+			stats.duration_per_type[type] += p->pkt_duration;
 	}
 }
 
@@ -178,11 +179,13 @@ update_spectrum(struct packet_info* p, struct node_info* n)
 	chan->durations += p->pkt_duration;
 	ewma_add(&chan->signal_avg, -chan->signal);
 
-	if (!n)
+	if (!n) {
+		DEBUG("spec no node\n");
 		return;
+	}
 
 	/* add node to channel if not already there */
-	list_for_each_entry(cn, &chan->nodes, chan_list) {
+	list_for_each(&chan->nodes, cn, chan_list) {
 		if (cn->node == n) {
 			DEBUG("SPEC node found %p\n", cn->node);
 			break;
@@ -194,8 +197,8 @@ update_spectrum(struct packet_info* p, struct node_info* n)
 		cn->node = n;
 		cn->chan = chan;
 		ewma_init(&cn->sig_avg, 1024, 8);
-		list_add_tail(&cn->chan_list, &chan->nodes);
-		list_add_tail(&cn->node_list, &n->on_channels);
+		list_add_tail(&chan->nodes, &cn->chan_list);
+		list_add_tail(&n->on_channels, &cn->node_list);
 		chan->num_nodes++;
 		n->num_on_channels++;
 	}
@@ -211,12 +214,12 @@ update_spectrum_durations(void)
 {
 	/* also if channel was not changed, keep stats only for every channel_time.
 	 * display code uses durations_last to get a more stable view */
-	if (conf.current_channel >= 0) {
-		spectrum[conf.current_channel].durations_last =
-				spectrum[conf.current_channel].durations;
-		spectrum[conf.current_channel].durations = 0;
-		ewma_add(&spectrum[conf.current_channel].durations_avg,
-			 spectrum[conf.current_channel].durations_last);
+	if (conf.channel_idx >= 0) {
+		spectrum[conf.channel_idx].durations_last =
+				spectrum[conf.channel_idx].durations;
+		spectrum[conf.channel_idx].durations = 0;
+		ewma_add(&spectrum[conf.channel_idx].durations_avg,
+			 spectrum[conf.channel_idx].durations_last);
 	}
 }
 
@@ -228,18 +231,21 @@ write_to_file(struct packet_info* p)
 		get_packet_type_name(p->wlan_type), ether_sprintf(p->wlan_src));
 	fprintf(DF, "%s, ", ether_sprintf(p->wlan_dst));
 	fprintf(DF, "%s, ", ether_sprintf(p->wlan_bssid));
-	fprintf(DF, "%x, %d, %d, %d, %d, %d, ",
+	fprintf(DF, "%x, %d, %d, %d, %d, %d, %d, ",
 		p->pkt_types, p->phy_signal, p->phy_noise, p->phy_snr,
-		p->wlan_len, p->phy_rate);
+		p->wlan_len, p->phy_rate, p->phy_freq);
 	fprintf(DF, "%016llx, ", (unsigned long long)p->wlan_tsf);
-	fprintf(DF, "%s, %d, %d, %d, ",
-		p->wlan_essid, p->wlan_mode, p->wlan_channel, p->wlan_wep);
+	fprintf(DF, "%s, %d, %d, %d, %d, %d, ",
+		p->wlan_essid, p->wlan_mode, p->wlan_channel,
+		p->wlan_wep, p->wlan_wpa, p->wlan_rsn);
 	fprintf(DF, "%s, ", ip_sprintf(p->ip_src));
 	fprintf(DF, "%s, ", ip_sprintf(p->ip_dst));
-	fprintf(DF, "%d, %d, %d\n", p->olsr_type, p->olsr_neigh, p->olsr_tc);
+	fprintf(DF, "%d, %d\n", p->olsr_type, p->olsr_neigh);
+	fflush(DF);
 }
 
 
+/* return 1 if packet is filtered */
 static int
 filter_packet(struct packet_info* p)
 {
@@ -248,7 +254,17 @@ filter_packet(struct packet_info* p)
 	if (conf.filter_off)
 		return 0;
 
-	if (!(p->pkt_types & conf.filter_pkt)) {
+	if (conf.filter_pkt != PKT_TYPE_ALL && (p->pkt_types & ~conf.filter_pkt)) {
+		stats.filtered_packets++;
+		return 1;
+	}
+
+	/* cannot trust anything if FCS is bad */
+	if (p->phy_flags & PHY_FLAG_BADFCS)
+		return 0;
+
+	if (conf.filter_mode != WLAN_MODE_ALL && ((p->wlan_mode & ~conf.filter_mode) || p->wlan_mode == 0)) {
+		/* this also filters out packets where we cannot associate a mode (ACK, RTS/CTS) */
 		stats.filtered_packets++;
 		return 1;
 	}
@@ -277,12 +293,12 @@ filter_packet(struct packet_info* p)
 void
 handle_packet(struct packet_info* p)
 {
-	struct node_info* n;
+	struct node_info* n = NULL;
 	int i = -1;
 
 	/* filter on server side only */
 	if (!conf.serveraddr && filter_packet(p)) {
-		if (!conf.quiet && !conf.paused)
+		if (!conf.quiet && !conf.paused && !DO_DEBUG)
 			update_display_clock();
 		return;
 	}
@@ -290,25 +306,28 @@ handle_packet(struct packet_info* p)
 	if (cli_fd != -1)
 		net_send_packet(p);
 
-	if (conf.dumpfile != NULL)
+	if (conf.dumpfile != NULL && !conf.paused && DF != NULL)
 		write_to_file(p);
 
-	if (conf.quiet || conf.paused)
+	if (conf.paused)
 		return;
 
+	DEBUG("handle %s\n", get_packet_type_name(p->wlan_type));
+
 	/* get channel index for packet */
-	if (p->phy_chan) {
-		/* find channel index from packet channel */
-		for (i = 0; i < conf.num_channels && i < MAX_CHANNELS; i++)
-			if (channels[i].chan == p->phy_chan)
-				break;
+	if (p->phy_freq) {
+		i = channel_find_index_from_freq(p->phy_freq);
 	}
 	/* not found from pkt, best guess from config but it might be
 	 * unknown (-1) too */
-	if (i < 0 || i >= conf.num_channels || i >= MAX_CHANNELS)
-		p->pkt_chan_idx = conf.current_channel;
+	if (i < 0)
+		p->pkt_chan_idx = conf.channel_idx;
 	else
 		p->pkt_chan_idx = i;
+	/* wlan_channel is only known for beacons and probe response,
+	 * otherwise we set it from the physical channel */
+	if (p->wlan_channel == 0 && p->pkt_chan_idx >= 0)
+		p->wlan_channel = channel_get_chan_from_idx(p->pkt_chan_idx);
 
 	/* detect if noise reading is present or not */
 	if (!conf.have_noise && p->phy_noise)
@@ -316,20 +335,25 @@ handle_packet(struct packet_info* p)
 
 	/* if current channel is unknown (this is a mac80211 bug), guess it from
 	 * the packet */
-	if (conf.current_channel < 0 && p->pkt_chan_idx >= 0)
-		conf.current_channel = p->pkt_chan_idx;
+	if (conf.channel_idx < 0 && p->pkt_chan_idx >= 0)
+		conf.channel_idx = p->pkt_chan_idx;
 
-	n = node_update(p);
+	if (!(p->phy_flags & PHY_FLAG_BADFCS)) {
+		/* we can't trust any fields except phy_* of packets with bad FCS,
+		 * so we can't do all this here */
+		n = node_update(p);
 
-	if (n)
-		p->wlan_retries = n->wlan_retries_last;
+		if (n)
+			p->wlan_retries = n->wlan_retries_last;
 
-	p->pkt_duration = ieee80211_frame_duration(
+		p->pkt_duration = ieee80211_frame_duration(
 				p->phy_flags & PHY_FLAG_MODE_MASK,
-				p->wlan_len, p->phy_rate * 5,
+				p->wlan_len, p->phy_rate,
 				p->phy_flags & PHY_FLAG_SHORTPRE,
-				0 /*shortslot*/, p->wlan_type, p->wlan_qos_class,
+				0 /*shortslot*/, p->wlan_type,
+				p->wlan_qos_class,
 				p->wlan_retries);
+	}
 
 	update_history(p);
 	update_statistics(p);
@@ -337,7 +361,8 @@ handle_packet(struct packet_info* p)
 	update_essids(p, n);
 
 #if !DO_DEBUG
-	update_display(p, n);
+	if (!conf.quiet)
+		update_display(p);
 #endif
 }
 
@@ -372,16 +397,20 @@ receive_any(void)
 	FD_ZERO(&read_fds);
 	FD_ZERO(&write_fds);
 	FD_ZERO(&excpt_fds);
-	FD_SET(0, &read_fds);
+	if (!conf.quiet && !DO_DEBUG)
+		FD_SET(0, &read_fds);
 	FD_SET(mon, &read_fds);
 	if (srv_fd != -1)
 		FD_SET(srv_fd, &read_fds);
 	if (cli_fd != -1)
 		FD_SET(cli_fd, &read_fds);
+	if (ctlpipe != -1)
+		FD_SET(ctlpipe, &read_fds);
 
 	tv.tv_sec = 0;
 	tv.tv_usec = min(conf.channel_time, 1000000);
 	mfd = max(mon, srv_fd);
+	mfd = max(mfd, ctlpipe);
 	mfd = max(mfd, cli_fd) + 1;
 
 	ret = select(mfd, &read_fds, &write_fds, &excpt_fds, &tv);
@@ -396,7 +425,7 @@ receive_any(void)
 		err(1, "select()");
 
 	/* stdin */
-	if (FD_ISSET(0, &read_fds) && !conf.quiet)
+	if (FD_ISSET(0, &read_fds) && !conf.quiet && !DO_DEBUG)
 		handle_user_input();
 
 	/* local packet or client */
@@ -414,6 +443,10 @@ receive_any(void)
 	/* from client to server */
 	if (cli_fd > -1 && FD_ISSET(cli_fd, &read_fds))
 		net_receive(cli_fd, cli_buffer, &cli_buflen, sizeof(cli_buffer));
+
+	/* named pipe */
+	if (ctlpipe > -1 && FD_ISSET(ctlpipe, &read_fds))
+		control_receive_command();
 }
 
 
@@ -426,14 +459,14 @@ free_lists(void)
 	struct chan_node *cn, *cn2;
 
 	/* free node list */
-	list_for_each_entry_safe(ni, mi, &nodes, list) {
+	list_for_each_safe(&nodes, ni, mi, list) {
 		DEBUG("free node %s\n", ether_sprintf(ni->last_pkt.wlan_src));
 		list_del(&ni->list);
 		free(ni);
 	}
 
 	/* free essids */
-	list_for_each_entry_safe(e, f, &essids.list, list) {
+	list_for_each_safe(&essids.list, e, f, list) {
 		DEBUG("free essid '%s'\n", e->essid);
 		list_del(&e->list);
 		free(e);
@@ -441,7 +474,7 @@ free_lists(void)
 
 	/* free channel nodes */
 	for (i = 0; i < conf.num_channels; i++) {
-		list_for_each_entry_safe(cn, cn2, &spectrum[i].nodes, chan_list) {
+		list_for_each_safe(&spectrum[i].nodes, cn, cn2, chan_list) {
 			DEBUG("free chan_node %p\n", cn);
 			list_del(&cn->chan_list);
 			cn->chan->num_nodes--;
@@ -459,8 +492,13 @@ finish_all(void)
 	if (!conf.serveraddr)
 		close_packet_socket(mon, conf.ifname);
 
-	if (DF != NULL)
+	if (DF != NULL) {
 		fclose(DF);
+		DF = NULL;
+	}
+
+	if (conf.allow_control)
+		control_finish();
 
 #if !DO_DEBUG
 	net_finish();
@@ -479,14 +517,14 @@ exit_handler(void)
 
 
 static void
-sigint_handler(int sig)
+sigint_handler(__attribute__((unused)) int sig)
 {
 	exit(0);
 }
 
 
 static void
-sigpipe_handler(int sig)
+sigpipe_handler(__attribute__((unused)) int sig)
 {
 	/* ignore signal here - we will handle it after write failed */
 }
@@ -498,7 +536,7 @@ get_options(int argc, char** argv)
 	int c;
 	static int n;
 
-	while((c = getopt(argc, argv, "hqsCi:t:c:p:e:d:o:b:")) > 0) {
+	while((c = getopt(argc, argv, "hqsCi:t:c:p:e:f:d:o:b:X::x:m:u:U:a:")) > 0) {
 		switch (c) {
 		case 'p':
 			conf.port = optarg;
@@ -538,22 +576,115 @@ get_options(int argc, char** argv)
 		case 'C':
 			conf.allow_client = 1;
 			break;
+		case 'X':
+			if (optarg != NULL)
+				conf.control_pipe = optarg;
+			conf.allow_control = 1;
+			break;
+		case 'x':
+			control_send_command(optarg);
+			exit(0);
+		case 'm':
+			if (conf.filter_mode == WLAN_MODE_ALL)
+				conf.filter_mode = 0;
+			if (strcmp(optarg, "AP") == 0)
+				conf.filter_mode |= WLAN_MODE_AP;
+			else if (strcmp(optarg, "STA") == 0)
+				conf.filter_mode |= WLAN_MODE_STA;
+			else if (strcmp(optarg, "ADH") == 0 || strcmp(optarg, "IBSS") == 0)
+				conf.filter_mode |= WLAN_MODE_IBSS;
+			else if (strcmp(optarg, "PRB") == 0)
+				conf.filter_mode |= WLAN_MODE_PROBE;
+			else if (strcmp(optarg, "WDS") == 0)
+				conf.filter_mode |= WLAN_MODE_4ADDR;
+			else if (strcmp(optarg, "UNKNOWN") == 0)
+				conf.filter_mode |= WLAN_MODE_UNKNOWN;
+			break;
+		case 'f':
+			if (conf.filter_pkt == PKT_TYPE_ALL)
+				conf.filter_pkt = 0;
+			if (strcmp(optarg, "CTRL") == 0 || strcmp(optarg, "CONTROL") == 0)
+				conf.filter_pkt |= PKT_TYPE_CTRL | PKT_TYPE_ALL_CTRL;
+			else if (strcmp(optarg, "MGMT") == 0 || strcmp(optarg, "MANAGEMENT") == 0)
+				conf.filter_pkt |= PKT_TYPE_MGMT | PKT_TYPE_ALL_MGMT;
+			else if (strcmp(optarg, "DATA") == 0)
+				conf.filter_pkt |= PKT_TYPE_DATA | PKT_TYPE_ALL_DATA;
+			else if (strcmp(optarg, "BADFCS") == 0)
+				conf.filter_pkt |= PKT_TYPE_BADFCS;
+			else if (strcmp(optarg, "BEACON") == 0)
+				conf.filter_pkt |= PKT_TYPE_BEACON;
+			else if (strcmp(optarg, "PROBE") == 0)
+				conf.filter_pkt |= PKT_TYPE_PROBE;
+			else if (strcmp(optarg, "ASSOC") == 0)
+				conf.filter_pkt |= PKT_TYPE_ASSOC;
+			else if (strcmp(optarg, "AUTH") == 0)
+				conf.filter_pkt |= PKT_TYPE_AUTH;
+			else if (strcmp(optarg, "RTS") == 0)
+				conf.filter_pkt |= PKT_TYPE_RTSCTS;
+			else if (strcmp(optarg, "ACK") == 0)
+				conf.filter_pkt |= PKT_TYPE_ACK;
+			else if (strcmp(optarg, "NULL") == 0)
+				conf.filter_pkt |= PKT_TYPE_NULL;
+			else if (strcmp(optarg, "QDATA") == 0)
+				conf.filter_pkt |= PKT_TYPE_QDATA;
+			else if (strcmp(optarg, "ARP") == 0)
+				conf.filter_pkt |= PKT_TYPE_ARP;
+			else if (strcmp(optarg, "IP") == 0)
+				conf.filter_pkt |= PKT_TYPE_IP;
+			else if (strcmp(optarg, "ICMP") == 0)
+				conf.filter_pkt |= PKT_TYPE_ICMP;
+			else if (strcmp(optarg, "UDP") == 0)
+				conf.filter_pkt |= PKT_TYPE_UDP;
+			else if (strcmp(optarg, "TCP") == 0)
+				conf.filter_pkt |= PKT_TYPE_TCP;
+			else if (strcmp(optarg, "OLSR") == 0)
+				conf.filter_pkt |= PKT_TYPE_OLSR;
+			else if (strcmp(optarg, "BATMAN") == 0)
+				conf.filter_pkt |= PKT_TYPE_BATMAN;
+			else if (strcmp(optarg, "MESHZ") == 0)
+				conf.filter_pkt |= PKT_TYPE_MESHZ;
+			/* if one of the individual subtype frames is selected we enable the general frame type */
+			if (conf.filter_pkt & PKT_TYPE_ALL_MGMT)
+				conf.filter_pkt |= PKT_TYPE_MGMT;
+			if (conf.filter_pkt & PKT_TYPE_ALL_CTRL)
+				conf.filter_pkt |= PKT_TYPE_CTRL;
+			if (conf.filter_pkt & PKT_TYPE_ALL_DATA)
+				conf.filter_pkt |= PKT_TYPE_DATA;
+			break;
 		case 'h':
 		default:
-			printf("usage: %s [-h] [-q] [-s] [-i interface] [-t sec] [-c IP] [-C] [-p port] [-e mac] [-d ms] [-o file] [-b bytes]\n\n"
-				"Options (default value)\n"
-				"  -h\t\tthis help\n"
-				"  -q\t\tquiet\n"
-				"  -s\t\tspectrum analyzer\n"
-				"  -i <intf>\tinterface (wlan0)\n"
-				"  -t <sec>\tnode timeout (60)\n"
-				"  -c <IP>\tconnect to server\n"
-				"  -C allow client connection (server)\n"
-				"  -p <port>\tuse port (4444)\n"
-				"  -e <mac>\tfilter all macs except these (multiple)\n"
-				"  -d <ms>\tdisplay update interval (100)\n"
-				"  -o <filename>\twrite packet info into file\n"
-				"  -b <bytes>\treceive buffer size (not set)\n"
+			printf("\nUsage: %s [-h] [-q] [-i interface] [-t sec] [-d ms] [-b bytes]\n"
+				"\t\t[-s] [-C] [-c IP] [-p port] [-o file] [-X[name]] [-x command]\n"
+				"\t\t[-e MAC] [-f PKT_NAME] [-m MODE]\n\n"
+
+				"General Options: Description (default value)\n"
+				"  -h\t\tHelp\n"
+				"  -q\t\tQuiet, no output\n"
+				"  -i <intf>\tInterface name (wlan0)\n"
+				"  -t <sec>\tNode timeout in seconds (60)\n"
+				"  -d <ms>\tDisplay update interval in ms (100)\n"
+				"  -b <bytes>\tReceive buffer size in bytes (not set)\n"
+
+				"\nFeature Options:\n"
+				"  -s\t\t(Poor mans) Spectrum analyzer mode\n\n"
+
+				"  -C\t\tAllow client connection, server mode (off)\n"
+				"  -c <IP>\tConnect to server with <IP>, client mode (off)\n"
+				"  -p <port>\tPort number of server (4444)\n\n"
+
+				"  -o <filename>\tWrite packet info into 'filename'\n\n"
+
+				"  -X[filename]\tAllow control socket on 'filename' (/tmp/horst)\n"
+				"  -x <command>\tSend control command\n"
+
+				"\nFilter Options:\n"
+				" Filters are generally 'positive' or 'inclusive' which means you define\n"
+				" what you want to see, and everything else is getting filtered out.\n"
+				" If a filter is not set it is inactive and nothing is filtered.\n"
+				" All filter options can be specified multiple times.\n"
+				"  -e <MAC>\tSource MAC addresses (xx:xx:xx:xx:xx:xx), up to 9 times\n"
+				"  -f <PKT_NAME>\tFilter packet types\n"
+				"  -m <MODE>\tOperating mode: AP|STA|ADH|PRB|WDS|UNKNOWN\n"
 				"\n",
 				argv[0]);
 			exit(0);
@@ -563,51 +694,67 @@ get_options(int argc, char** argv)
 }
 
 
+void
+init_spectrum(void)
+{
+	int i;
+
+	for (i = 0; i < conf.num_channels && i < MAX_CHANNELS; i++) {
+		list_head_init(&spectrum[i].nodes);
+		ewma_init(&spectrum[i].signal_avg, 1024, 8);
+		ewma_init(&spectrum[i].durations_avg, 1024, 8);
+	}
+}
+
+
 int
 main(int argc, char** argv)
 {
-	INIT_LIST_HEAD(&essids.list);
-	INIT_LIST_HEAD(&nodes);
+	list_head_init(&essids.list);
+	list_head_init(&nodes);
 
 	get_options(argc, argv);
 
 	signal(SIGINT, sigint_handler);
+	signal(SIGTERM, sigint_handler);
+	signal(SIGHUP, sigint_handler);
 	signal(SIGPIPE, sigpipe_handler);
 	atexit(exit_handler);
 
 	gettimeofday(&stats.stats_time, NULL);
 	gettimeofday(&the_time, NULL);
 
-	conf.current_channel = -1;
+	conf.channel_idx = -1;
+
+	if (conf.allow_control) {
+		printlog("Allowing control socket");
+		control_init_pipe();
+	}
 
 	if (conf.serveraddr)
 		mon = net_open_client_socket(conf.serveraddr, conf.port);
 	else {
-		mon = open_packet_socket(conf.ifname, sizeof(buffer), conf.recv_buffer_size);
+		mon = open_packet_socket(conf.ifname, conf.recv_buffer_size);
 		if (mon <= 0)
 			err(1, "Couldn't open packet socket");
 
-		conf.arphrd = device_get_arptype(mon, conf.ifname);
+		conf.arphrd = device_get_hwinfo(mon, conf.ifname, conf.my_mac_addr);
 		if (conf.arphrd != ARPHRD_IEEE80211_PRISM &&
-		conf.arphrd != ARPHRD_IEEE80211_RADIOTAP) {
-			printf("Wrong monitor type! Please use radiotap or prism2 headers\n");
+		    conf.arphrd != ARPHRD_IEEE80211_RADIOTAP) {
+			printf("You need to put your interface into monitor mode!\n");
+			printf("(e.g. 'iw %s interface add mon0 type monitor' and 'horst -i mon0')\n", conf.ifname);
 			exit(1);
 		}
 
-		/* get available channels */
-		conf.num_channels = wext_get_channels(mon, conf.ifname, channels);
-		init_channels();
-		get_current_channel(mon);
+		channel_init();
+		init_spectrum();
 	}
 
 	if (!conf.quiet && !DO_DEBUG)
 		init_display();
 
-	if (conf.dumpfile != NULL) {
-		DF = fopen(conf.dumpfile, "w");
-		if (DF == NULL)
-			err(1, "Couldn't open dump file");
-	}
+	if (conf.dumpfile != NULL)
+		dumpfile_open(conf.dumpfile);
 
 	if (!conf.serveraddr && conf.port && conf.allow_client)
 		net_init_server_socket(conf.port);
@@ -617,11 +764,13 @@ main(int argc, char** argv)
 		receive_any();
 		gettimeofday(&the_time, NULL);
 		timeout_nodes();
+
 		if (!conf.serveraddr) { /* server */
-			if (auto_change_channel(mon)) {
+			if (channel_auto_change()) {
 				net_send_channel_config();
 				update_spectrum_durations();
-				update_display(NULL, NULL);
+				if (!conf.quiet && !DO_DEBUG)
+					update_display(NULL);
 			}
 		}
 	}
@@ -630,6 +779,37 @@ main(int argc, char** argv)
 }
 
 
+void
+main_pause(int pause)
+{
+	conf.paused = pause;
+	printlog(conf.paused ? "- PAUSED -" : "- RESUME -");
+}
+
+
+void
+dumpfile_open(char* name)
+{
+	if (DF != NULL) {
+		fclose(DF);
+		DF = NULL;
+	}
+
+	if (name == NULL || strlen(name) == 0) {
+		printlog("- Not writing outfile");
+		conf.dumpfile = NULL;
+		return;
+	}
+
+	conf.dumpfile = name;
+	DF = fopen(conf.dumpfile, "w");
+	if (DF == NULL)
+		err(1, "Couldn't open dump file");
+
+	printlog("- Writing to outfile %s", conf.dumpfile);
+}
+
+
 #if 0
 void print_rate_duration_table(void)
 {
